﻿from google.colab import files
import os
import time,datetime
import numpy as np
import cv2
EPS = 1e-15
INF = 1e+38

def main():
	"""
	Main process

	Parameters
	----------
	None

	Returns
	-------
	None
	"""

	# Read input image
	print('Please upload 1 image file')
	uploaded=files.upload()
	filename=list(uploaded.keys())[0]
	print(datetime.datetime.fromtimestamp(time.time()).strftime('%Y/%m/%d %H:%M:%S'))
	print('Read input image')
	if not os.path.exists(filename):
		print('File not exist:', filename)
		return
	image = cv2.imread(filename, cv2.IMREAD_COLOR)
	if image is None:
		print('Cannot read file:', filename)
		return
	rows,cols,_=image.shape

	# Calculate XYZ
	print(datetime.datetime.fromtimestamp(time.time()).strftime('%Y/%m/%d %H:%M:%S'))
	print('Calculate XYZ')
	largexyz=np.zeros((rows,cols,3),dtype=np.float64)
	for y in range(rows):
		for x in range(cols):
			colr = image[y, x, 2]
			colg = image[y, x, 1]
			colb = image[y, x, 0]
			colx, coly, colz = ConvertRGB2XYZ(colr, colg, colb)
			largexyz[y, x, 2] = colx
			largexyz[y, x, 1] = coly
			largexyz[y, x, 0] = colz

	# Calculate P-dichromat
	print(datetime.datetime.fromtimestamp(time.time()).strftime('%Y/%m/%d %H:%M:%S'))
	print('Calculate P-dichromat')
	simulatep = np.zeros((rows,cols,3), dtype=np.uint8)
	for y in range(0, rows):
		for x in range(0, cols):
			colx = largexyz[y, x, 2]
			coly = largexyz[y, x, 1]
			colz = largexyz[y, x, 0]
			colx, coly, colz = Convert2BlindP(colx, coly, colz)
			colr, colg, colb = ConvertXYZ2RGB(colx, coly, colz)
			simulatep[y, x, 2] = (int)(colr + 0.5)
			simulatep[y, x, 1] = (int)(colg + 0.5)
			simulatep[y, x, 0] = (int)(colb + 0.5)

	# Calculate D-dichromat
	print(datetime.datetime.fromtimestamp(time.time()).strftime('%Y/%m/%d %H:%M:%S'))
	print('Calculate D-dichromat')
	simulated = np.zeros((rows,cols,3), dtype=np.uint8)
	for y in range(0, rows):
		for x in range(0, cols):
			colx = largexyz[y, x, 2]
			coly = largexyz[y, x, 1]
			colz = largexyz[y, x, 0]
			colx, coly, colz = Convert2BlindD(colx, coly, colz)
			colr, colg, colb = ConvertXYZ2RGB(colx, coly, colz)
			simulated[y, x, 2] = (int)(colr + 0.5)
			simulated[y, x, 1] = (int)(colg + 0.5)
			simulated[y, x, 0] = (int)(colb + 0.5)

	# Calculate xy
	print(datetime.datetime.fromtimestamp(time.time()).strftime('%Y/%m/%d %H:%M:%S'))
	print('Calculate xy')
	smallxyz = np.zeros((rows, cols, 3), dtype=np.float64)
	for y in range(0, rows):
		for x in range(0, cols):
			colx = largexyz[y, x, 2]
			coly = largexyz[y, x, 1]
			colz = largexyz[y, x, 0]

			chromax = colx / (colx + coly + colz + EPS)
			chromay = coly / (colx + coly + colz + EPS)
			chromaz = colz / (colx + coly + colz + EPS)

			smallxyz[y, x, 2] = chromax
			smallxyz[y, x, 1] = chromay
			smallxyz[y, x, 0] = chromaz

	# Calculate hue
	print(datetime.datetime.fromtimestamp(time.time()).strftime('%Y/%m/%d %H:%M:%S'))
	print('Calculate hue')
	betanew = np.zeros((rows, cols), dtype=np.float64)
	for y in range(0, rows):
		for x in range(0, cols):
			chromax = smallxyz[y, x, 2]
			chromay = smallxyz[y, x, 1]
			beta = np.arctan2(chromay - 0.33, chromax - 0.33)
			beta = np.degrees(beta)
			if beta < 360: beta += 360
			beta += 45
			if beta > 360: beta -= 360
			betanew[y, x] = beta

	# Calculate histogram
	print(datetime.datetime.fromtimestamp(time.time()).strftime('%Y/%m/%d %H:%M:%S'))
	print('Calculate histogram')
	betamin = np.min(betanew)
	betamax = np.max(betanew)
	histogram = [0 for _ in range(1000)]
	for y in range(rows):
		for x in range(cols):
			beta = betanew[y, x]
			index = (beta - betamin) / (betamax - betamin) * 1000
			index = np.clip(index, 0, 999)
			histogram[int(index)] += 1
	filedata = [str(i) + '\n' for i in histogram]
	with open('histogram.txt', mode='w') as f:
		f.writelines(filedata)
	for i in range(1, 1000):
		histogram[i] = histogram[i] + histogram[i - 1]

	# Enhanced image: P-dichromat
	print(datetime.datetime.fromtimestamp(time.time()).strftime('%Y/%m/%d %H:%M:%S'))
	print('Enhanced image: P-dichromat')
	enhancep = np.zeros((rows, cols, 3), dtype=np.uint8)
	enhanceprgb = np.zeros((rows, cols, 3), dtype=np.uint8)
	for y in range(rows):
		for x in range(cols):
			beta = betanew[y, x]
			index = (beta - betamin) / (betamax - betamin) * 1000
			index = np.clip(index, 0, 999)
			ratio = histogram[int(index)] / histogram[999]
			beta = ratio * (200 - 140) + 140

			colx = smallxyz[y, x, 2]
			coly = smallxyz[y, x, 1]
			colz = smallxyz[y, x, 0]
			beta = np.radians(beta)
			length = (colx - 0.747) * np.cos(beta) + (coly - 0.278) * np.sin(beta)
			colx = np.cos(beta) * length + 0.747
			coly = np.sin(beta) * length + 0.278
			colz = 1 - colx - coly

			origr = image[y, x, 2]
			origg = image[y, x, 1]
			origb = image[y, x, 0]
			colr, colg, colb = ConvertXYZ2RGB(colx, coly, colz, origr, origg, origb)
			enhanceprgb[y, x, 2] = int(colr + 0.5)
			enhanceprgb[y, x, 1] = int(colg + 0.5)
			enhanceprgb[y, x, 0] = int(colb + 0.5)

			colx, coly, colz = ConvertRGB2XYZ(colr, colg, colb)
			colx, coly, colz = Convert2BlindP(colx, coly, colz)
			colr, colg, colb = ConvertXYZ2RGB(colx, coly, colz)
			enhancep[y, x, 2] = int(colr + 0.5)
			enhancep[y, x, 1] = int(colg + 0.5)
			enhancep[y, x, 0] = int(colb + 0.5)

	# Enhanced image: D-dichromat
	print(datetime.datetime.fromtimestamp(time.time()).strftime('%Y/%m/%d %H:%M:%S'))
	print('Enhanced image: D-dichromat')
	enhanced = np.zeros((rows, cols, 3), dtype=np.uint8)
	enhancedrgb = np.zeros((rows, cols, 3), dtype=np.uint8)
	for y in range(rows):
		for x in range(cols):
			beta = betanew[y, x]
			index = (beta - betamin) / (betamax - betamin) * 1000
			index = np.clip(index, 0, 999)
			ratio = histogram[int(index)] / histogram[999]
			beta = ratio * (170 - 140) + 140

			colx = smallxyz[y, x, 2]
			coly = smallxyz[y, x, 1]
			colz = smallxyz[y, x, 0]
			beta = np.radians(beta)
			length = (colx - 1.000) * np.cos(beta) + (coly - 0.000) * np.sin(beta)
			colx = np.cos(beta) * length + 1.000
			coly = np.sin(beta) * length + 0.000
			colz = 1 - colx - coly

			origr = image[y, x, 2]
			origg = image[y, x, 1]
			origb = image[y, x, 0]
			colr, colg, colb = ConvertXYZ2RGB(colx, coly, colz, origr, origg, origb)
			enhancedrgb[y, x, 2] = int(colr + 0.5)
			enhancedrgb[y, x, 1] = int(colg + 0.5)
			enhancedrgb[y, x, 0] = int(colb + 0.5)

			colx, coly, colz = ConvertRGB2XYZ(colr, colg, colb)
			colx, coly, colz = Convert2BlindP(colx, coly, colz)
			colr, colg, colb = ConvertXYZ2RGB(colx, coly, colz)
			enhanced[y, x, 2] = int(colr + 0.5)
			enhanced[y, x, 1] = int(colg + 0.5)
			enhanced[y, x, 0] = int(colb + 0.5)

	# Output results
	print(datetime.datetime.fromtimestamp(time.time()).strftime('%Y/%m/%d %H:%M:%S'))
	print('Output results')
	cv2.imwrite('simulated.bmp', simulated)
	cv2.imwrite('simulatep.bmp', simulatep)
	cv2.imwrite('enhanced.bmp', enhanced)
	cv2.imwrite('enhancep.bmp', enhancep)
	cv2.imwrite('enhancedrgb.bmp', enhancedrgb)
	cv2.imwrite('enhanceprgb.bmp', enhanceprgb)
	files.download('simulated.bmp')
	files.download('simulatep.bmp')
	files.download('enhanced.bmp')
	files.download('enhancep.bmp')
	files.download('enhancedrgb.bmp')
	files.download('enhanceprgb.bmp')

	# Finish
	print(datetime.datetime.fromtimestamp(time.time()).strftime('%Y/%m/%d %H:%M:%S'))
	print('Done')

def ConvertRGB2XYZ(colr, colg, colb):
	"""
	Convert sRGB to XYZ

	Parameters
	----------
	colr : float
	colg : float
	colb : float

	Returns
	-------
	colx : float
	coly : float
	colz : float
	"""

	colr /= 255
	colg /= 255
	colb /= 255
	if colr <= 0.04045:	colr /= 12.92
	else:				colr = ((colr + 0.055) / 1.055) ** 2.4
	if colg <= 0.04045:	colg /= 12.92
	else:				colg = ((colg + 0.055) / 1.055) ** 2.4
	if colb <= 0.04045:	colb /= 12.92
	else:				colb = ((colb + 0.055) / 1.055) ** 2.4
	colx = 0.4124 * colr + 0.3576 * colg + 0.1805 * colb
	coly = 0.2126 * colr + 0.7152 * colg + 0.0722 * colb
	colz = 0.0193 * colr + 0.1192 * colg + 0.9505 * colb
	return (colx, coly, colz)

def ConvertXYZ2RGB(colx, coly, colz, origr=None, origg=None, origb=None):
	"""
	Convert XYZ to sRGB

	Parameters
	----------
	colx : float
	coly : float
	colz : float
	origr : float
	origg : float
	origb : float

	Returns
	-------
	colr : float
	colg : float
	colb : float
	"""

	colr =  3.2406 * colx - 1.5372 * coly - 0.4986 * colz
	colg = -0.9689 * colx + 1.8758 * coly + 0.0415 * colz
	colb =  0.0557 * colx - 0.204  * coly + 1.057  * colz
	if colr <= 0: colr = 0
	if colg <= 0: colg = 0
	if colb <= 0: colb = 0
	if colr > 0.0031308:	colr = 1.055 * (colr ** (1 / 2.4)) - 0.055
	else:					colr = 12.92 * colr
	if colg > 0.0031308:	colg = 1.055 * (colg ** (1 / 2.4)) - 0.055
	else:					colg = 12.92 * colg
	if colb > 0.0031308:	colb = 1.055 * (colb ** (1 / 2.4)) - 0.055
	else:					colb = 12.92 * colb
	if origr is None or origg is None or origb is None:
		colr *= 255
		colg *= 255
		colb *= 255
	else:
		scale = (float(origr) + float(origg) + float(origb)) / (colr + colg + colb + EPS)
		colr *= scale
		colg *= scale
		colb *= scale
	if colr >= 255: colr = 255
	if colg >= 255: colg = 255
	if colb >= 255: colb = 255
	return (colr, colg, colb)


def Convert2BlindP(colx, coly, colz):
	"""
	Convert to P-dichromat

	Parameters
	----------
	colx : float
	coly : float
	colz : float

	Returns
	-------
	colx : float
	coly : float
	colz : float
	"""

	# Convert XYZ to LMS
	coll =  0.155 * colx + 0.543 * coly - 0.0031 * colz
	colm = -0.155 * colx + 0.457 * coly + 0.0329 * colz
	cols =                                0.016  * colz

	# Convert to P-dichromat
	collp = 2.02 * colm - 2.52 * cols
	colmp = colm
	colsp = cols

	# Convert LMS to XYZ
	colx = 2.9484 * collp - 3.5032 * colmp + 7.7748 * colsp
	coly =          collp +          colmp - 1.8625 * colsp
	colz =                                  62.5    * colsp

	return (colx, coly, colz)

def Convert2BlindD(colx, coly, colz):
	"""
	Convert to D-dichromat

	Parameters
	----------
	colx : float
	coly : float
	colz : float

	Returns
	-------
	colx : float
	coly : float
	colz : float
	"""

	# Convert XYZ to LMS
	coll =  0.155 * colx + 0.543 * coly - 0.0031 * colz
	colm = -0.155 * colx + 0.457 * coly + 0.0329 * colz
	cols =                                0.016  * colz

	# Convert to D-dichromat
	colld = coll
	colmd = 0.49 * coll + 1.25 * cols
	colsd = cols

	# Convert LMS to XYZ
	colx = 2.9484 * colld - 3.5032 * colmd + 7.7748 * colsd
	coly =          colld +          colmd - 1.8625 * colsd
	colz =                                  62.5    * colsd

	return (colx, coly, colz)

if __name__ == '__main__':
	main()
